/* 
   Librairie mcimage : 

   fonctions pour les entrees/sortie fichiers et l'allocation de structures
   image en memoire.

   Michel Couprie
*/

#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <mcimage.h>

#define BUFFERSIZE 4096

/* ==================================== */
struct xvimage *allocimage(
  char * name,
  int32_t rs,   /* row size */
  int32_t cs,   /* col size */
  int32_t d,    /* depth */
  int32_t t)    /* data type */
/* ==================================== */
#undef F_NAME
#define F_NAME "allocimage"
{
  int32_t N = rs * cs * d;             /* taille image */
  struct xvimage *g;
  int32_t ts;                          /* type size */

  switch (t)
  {
    case VFF_TYP_BIT:      ts = 1; break;
    case VFF_TYP_1_BYTE:   ts = 1; break;
    case VFF_TYP_2_BYTE:   ts = 2; break;
    case VFF_TYP_4_BYTE:   ts = 4; break;
    case VFF_TYP_FLOAT:    ts = sizeof(float); break;
    case VFF_TYP_DOUBLE:   ts = sizeof(double); break;
    default: fprintf(stderr,"%s : bad data type %d\n", F_NAME, t);
             return NULL;
  } /* switch (t) */

  g = (struct xvimage *)malloc(sizeof(struct xvimage));
  if (g == NULL)
  {   
    fprintf(stderr,"%s : malloc failed (%d bytes)\n", F_NAME, sizeof(struct xvimage));
    return NULL;
  }

  g->image_data = (void *)calloc(1, N * ts);
  if (g == NULL)
  {   
    fprintf(stderr,"%s : calloc failed (%d bytes)\n", F_NAME, N * ts);
    return NULL;
  }

  if (name != NULL)
  {
    g->name = (char *)calloc(1,strlen(name)+1);
    if (g->name == NULL)
    {   fprintf(stderr,"%s : malloc failed for name\n", F_NAME);
        return NULL;
    }
    strcpy((char *)(g->name), name);
  }
  else
    g->name = NULL;

  rowsize(g) = rs;
  colsize(g) = cs;
  depth(g) = d;
  datatype(g) = t;
  g->xdim = g->ydim = g->zdim = 0.0;

  return g;
} /* allocimage() */

/* ==================================== */
void freeimage(struct xvimage *image)
/* ==================================== */
{
  free(image->image_data);
  free(image);
}

/* ==================================== */
struct xvimage *copyimage(struct xvimage *f)
/* ==================================== */
#undef F_NAME
#define F_NAME "copyimage"
{
  int32_t rs = rowsize(f);         /* taille ligne */
  int32_t cs = colsize(f);         /* taille colonne */
  int32_t ds = depth(f);           /* nb plans */
  int32_t N = rs * cs * ds;        /* taille image */
  int32_t type = datatype(f);
  struct xvimage *g;

  g = allocimage(NULL, rs, cs, ds, type);
  if (g == NULL)
  {
    fprintf(stderr,"%s : allocimage failed\n", F_NAME);
    return NULL;
  }

  switch(type)
  {
    case VFF_TYP_1_BYTE: memcpy(g->image_data, f->image_data, (N*sizeof(int8_t))); break;
    case VFF_TYP_4_BYTE: memcpy(g->image_data, f->image_data, (N*sizeof(int32_t))); break;
    case VFF_TYP_FLOAT:  memcpy(g->image_data, f->image_data, (N*sizeof(float))); break;
    case VFF_TYP_DOUBLE: memcpy(g->image_data, f->image_data, (N*sizeof(double))); break;
    default:
      fprintf(stderr,"%s : bad data type %d\n", F_NAME, type);
      return NULL;
  } /* switch(f->datatype) */

  if (f->name != NULL)
  {
    g->name = (char *)calloc(1,strlen(f->name) + 1);
    if (g->name == NULL)
    {   fprintf(stderr,"%s : malloc failed for name\n", F_NAME);
        return NULL;
    }
    strcpy(g->name, f->name);
  }
  return g;
} // copyimage()

/* ==================================== */
void writeimage(struct xvimage * image, char *filename)
/* ==================================== */
#undef F_NAME
#define F_NAME "writeimage"
{
  writeascimage(image, filename); 
} /* writeimage() */

/* ==================================== */
void writeascimage(struct xvimage * image, char *filename)
/* ==================================== */
#undef F_NAME
#define F_NAME "writeascimage"
{
  FILE *fd = NULL;
  int32_t rs, cs, ps, d, nndg, N, i;

  fd = fopen(filename,"w");
  if (!fd)
  {
    fprintf(stderr, "%s: cannot open file: %s\n", F_NAME, filename);
    exit(0);
  }

  rs = rowsize(image);
  cs = colsize(image);
  d = depth(image);
  ps = rs * cs;
  N = ps * d;

  if (datatype(image) == VFF_TYP_1_BYTE)
  {
    fputs("P2\n", fd);
    if ((image->xdim != 0.0) && (d > 1))
      fprintf(fd, "#xdim %g\n#ydim %g\n#zdim %g\n", image->xdim, image->ydim, image->zdim);
    if (d > 1) fprintf(fd, "%d %d %d\n", rs, cs, d); else  fprintf(fd, "%d %d\n", rs, cs);
    fprintf(fd, "255\n");

    for (i = 0; i < N; i++)
    {
      if (i % rs == 0) fprintf(fd, "\n");
      if (i % ps == 0) fprintf(fd, "\n");
      fprintf(fd, "%3d ", (int32_t)(UCHARDATA(image)[i]));
    } /* for i */
    fprintf(fd, "\n");
  }
  else if (datatype(image) == VFF_TYP_4_BYTE)
  {
    fputs("PB\n", fd);
    if ((image->xdim != 0.0) && (d > 1))
      fprintf(fd, "#xdim %g\n#ydim %g\n#zdim %g\n", image->xdim, image->ydim, image->zdim);
    if (d > 1) fprintf(fd, "%d %d %d\n", rs, cs, d); else  fprintf(fd, "%d %d\n", rs, cs);
    fprintf(fd, "4294967295\n");

    for (i = 0; i < N; i++)
    {
      if (i % rs == 0) fprintf(fd, "\n");
      if (i % ps == 0) fprintf(fd, "\n");
      fprintf(fd, "%ld ", ULONGDATA(image)[i]);
    } /* for i */
    fprintf(fd, "\n");
  }
  else if (datatype(image) == VFF_TYP_FLOAT)
  {
    fputs("PA\n", fd);
    if ((image->xdim != 0.0) && (d > 1))
      fprintf(fd, "#xdim %g\n#ydim %g\n#zdim %g\n", image->xdim, image->ydim, image->zdim);
    if (d > 1) fprintf(fd, "%d %d %d\n", rs, cs, d); else  fprintf(fd, "%d %d\n", rs, cs);
    fprintf(fd, "1\n");

    for (i = 0; i < N; i++)
    {
      if (i % rs == 0) fprintf(fd, "\n");
      if (i % ps == 0) fprintf(fd, "\n");
      fprintf(fd, "%8g ", FLOATDATA(image)[i]);
    } /* for i */
    fprintf(fd, "\n");
  }
  fclose(fd);
}

/* ==================================== */
void printimage(struct xvimage * image)
/* ==================================== */
#undef F_NAME
#define F_NAME "printimage"
{
  int32_t rs, cs, d, ps, N, i;

  rs = rowsize(image);
  cs = colsize(image);
  d = depth(image);
  ps = rs * cs;
  N = ps * d;

  for (i = 0; i < N; i++)
  {
    if (i % rs == 0) printf("\n");
    if (i % ps == 0) printf("\n");
    printf("%3d ", (int32_t)(UCHARDATA(image)[i]));
  } /* for i */
  printf("\n");
}

/* ==================================== */
struct xvimage * readimage(char *filename)
/* ==================================== */
#undef F_NAME
#define F_NAME "readimage"
{
  char buffer[BUFFERSIZE];
  FILE *fd = NULL;
  int32_t rs, cs, d, ndgmax, N, i;
  struct xvimage * image;
  int32_t ascii;  
  int32_t typepixel;
  int32_t c;
  double xdim=1.0, ydim=1.0, zdim=1.0;
  char *read;

  fd = fopen(filename,"r");
  if (!fd)
  {
    fprintf(stderr, "%s: file not found: %s\n", F_NAME, filename);
    return NULL;
  }

  read = fgets(buffer, BUFFERSIZE, fd); /* P5: raw byte bw  ; P2: ascii bw */
                                 /* P6: raw byte rgb ; P3: ascii rgb */
                                 /* P8: raw int32_t 2d-3d  ==  extension MC */
                                 /* P9: raw float 2d-3d  ==  extension MC */
                                 /* PA: ascii float 2d-3d  ==  extension LN */
                                 /* PB: ascii int32_t 2d-3d  ==  extension MC */

                                 /* P7: raw byte 3d : OBSOLETE - left for compatibility */
  if (!read)
  {
    fprintf(stderr, "%s: fgets returned without reading\n", F_NAME);
    return 0;
  }

  if (buffer[0] != 'P')
  {   fprintf(stderr,"%s : invalid image format\n", F_NAME);
      return NULL;
  }
  switch (buffer[1])
  {
    case '2': ascii = 1; typepixel = VFF_TYP_1_BYTE; break;
    case '5':
    case '7': ascii = 0; typepixel = VFF_TYP_1_BYTE; break;
    case '8': ascii = 0; typepixel = VFF_TYP_4_BYTE; break;
    case '9': ascii = 0; typepixel = VFF_TYP_FLOAT; break;
    case 'A': ascii = 1; typepixel = VFF_TYP_FLOAT; break;
    case 'B': ascii = 1; typepixel = VFF_TYP_4_BYTE; break;
    default:
      fprintf(stderr,"%s : invalid image format : P%c\n", F_NAME, buffer[1]);
      return NULL;
  } /* switch */

  do
  {
    read = fgets(buffer, BUFFERSIZE, fd); /* commentaire */
    if (!read)
    {
      fprintf(stderr, "%s: fgets returned without reading\n", F_NAME);
      return 0;
    }
    if (strncmp(buffer, "#xdim", 5) == 0)
      sscanf(buffer+5, "%lf", &xdim);
    else if (strncmp(buffer, "#ydim", 5) == 0)
      sscanf(buffer+5, "%lf", &ydim);
    else if (strncmp(buffer, "#zdim", 5) == 0)
      sscanf(buffer+5, "%lf", &zdim);
  } while (!isdigit(buffer[0]));

  c = sscanf(buffer, "%d %d %d", &rs, &cs, &d);
  if (c == 2) d = 1;
  else if (c != 3)
  {   fprintf(stderr,"%s : invalid image format\n", F_NAME);
      return NULL;
  }

  read = fgets(buffer, BUFFERSIZE, fd);
  if (!read)
  {
    fprintf(stderr, "%s: fgets returned without reading\n", F_NAME);
    return 0;
  }

  sscanf(buffer, "%d", &ndgmax);
  N = rs * cs * d;

  image = allocimage(NULL, rs, cs, d, typepixel);
  if (image == NULL)
  {   fprintf(stderr,"%s : alloc failed\n", F_NAME);
      return(NULL);
  }
  image->xdim = xdim;
  image->ydim = ydim;
  image->zdim = zdim;

  if (typepixel == VFF_TYP_1_BYTE)
  {
    if (ascii)
    {
      if (ndgmax == 255)
        for (i = 0; i < N; i++)
        {
          fscanf(fd, "%d", &c);
          UCHARDATA(image)[i] = (uint8_t)c;
        } /* for i */
      else if (ndgmax == 65535)
        for (i = 0; i < N; i++)
        {
          fscanf(fd, "%d", &c);
          UCHARDATA(image)[i] = (uint8_t)(c/256);
        } /* for i */
      else
      {
        fprintf(stderr,"%s : wrong ndgmax = %d\n", F_NAME, ndgmax);
        return(NULL);
      }
    }
    else
    {
      int32_t ret = fread(UCHARDATA(image), sizeof(char), N, fd);
      if (ret != N)
      {
        fprintf(stderr,"%s : fread failed : %d asked ; %d read\n", F_NAME, N, ret);
        return(NULL);
      }
    }
  } /* if (typepixel == VFF_TYP_1_BYTE) */
  else
  if (typepixel == VFF_TYP_4_BYTE)
  {
    if (ascii)
    {
      for (i = 0; i < N; i++)
      {
        fscanf(fd, "%ld", &(ULONGDATA(image)[i]));
      } /* for i */
    }
    else 
    {
      int32_t ret = fread(ULONGDATA(image), sizeof(int32_t), N, fd);
      if (ret != N)
      {
        fprintf(stderr,"%s : fread failed : %d asked ; %d read\n", F_NAME, N, ret);
        return(NULL);
      }
    }
  } /* if (typepixel == VFF_TYP_4_BYTE) */
  else
  if (typepixel == VFF_TYP_FLOAT)
  {
    if (ascii)
    {
      for (i = 0; i < N; i++)
      {
        fscanf(fd, "%f", &(FLOATDATA(image)[i]));
      } /* for i */
    }
    else 
    {
      int32_t ret = fread(FLOATDATA(image), sizeof(float), N, fd);
      if (ret != N)
      {
        fprintf(stderr,"%s : fread failed : %d asked ; %d read\n", F_NAME, N, ret);
        return(NULL);
      }
    }
  } /* if (typepixel == VFF_TYP_FLOAT) */

  fclose(fd);

  return image;
} /* readimage() */
